/* --------------------------------
              Gc.c
    Garbage Collection DYL code
        Subclasses AIDLE
   Dan Ramage - March 15, 1997
---------------------------------*/

#include <java.g>

#define _gc ((PR_GC *)DatApp1)

LOCAL_C VOID CollectGarbage(VOID)
/*  Destroy unused objects  */
    {
    OBJECTS *jro;
    UINT pos;

    if (_gc->gc.jrocount == NULL)
        return;

    for (pos = 0; pos < _gc->gc.jrocount; pos++)
        {
        jro = (OBJECTS *)(_gc->gc.jros + sizeof(OBJECTS) * pos);
        if (jro->count <= 0)
            {
            p_send2(jro->jro, O_DESTROY);
            /*  Finalize and destroy object  */
            p_adjust(_gc->gc.jros, sizeof(OBJECTS) * pos, -sizeof(OBJECTS));
            /*  Remove gap from heap data structure  */
            }
        }
    }

#pragma METHOD_CALL

METHOD VOID gc_destroy(PR_GC *self)
    {
    if (self->gc.jros)
        p_free(self->gc.jros);
    }

METHOD VOID gc_ao_init(PR_GC *self)
    {
#pragma warn(wpcv=>off)
    p_bcpy(DatApp1, &self, 2);
    /*  Put a copy of our handle at DatApp1  */
#pragma warn(wpcv=>on)
    self->gc.jros = NULL;
    self->gc.jropos = NULL;
    self->gc.jrocount = NULL;
    p_supersend2(self, O_AO_INIT);
    }

METHOD VOID gc_ao_run(PR_GC *self)
/*  Go through our list of objects giving up to GC_CALLS_PER_PASS jro object
threads or currently executing non-thread objects a chance to run.  Then,  
every GC_GARBAGE_COLLECT time through the loop, we garbage collect  */
    {
    OBJECTS *jro;
    UBYTE pos;

    if (self->gc.jrocount == NULL)
        return;

    for (pos = 0; pos < GC_CALLS_PER_PASS;)
        {
        if (++self->gc.jropos > self->gc.jrocount);
            self->gc.jropos -= self->gc.jrocount;
        
        jro = (OBJECTS *)(self->gc.jros+sizeof(OBJECTS)*self->gc.jropos);
        
        if ((jro->flags & GC_JRO_RUN) || (jro->flags & GC_JRO_THREAD))
        /*  Run specified jro if it is active or a thread  */
            {
            p_send2(jro->jro, O_AO_RUN);
            pos++;
            }
        }

    if ((++self->gc.collect % GC_GARBAGE_COLLECT) == FALSE)
    /*  Get rid of objects no longer in use  */
        CollectGarbage();
    }

METHOD UINT gc_gc_new_ref(PR_GC *self, TEXT *classname)
/*  Create a new JRO and add it to the jros struct and return its symbolic  
handle  */
    {
    /*  Real code goes here...  (Needs class loader before continuing)  */
    return NULL;
    }

METHOD VOID gc_gc_send_mess(PR_GC *self, UINT handle, TEXT *method)
/*  Send a message to an object via a symbolic handle  */
    {
    OBJECTS *jro;

    jro = (OBJECTS *)(self->gc.jros+sizeof(OBJECTS)*handle);
    p_send3(jro, O_JRO_SEND_MESS, method);
    }

METHOD UINT gc_gc_dec_access(PR_GC *self, UINT handle)
/*  Decrement access count of object 'handle'  */
    {
    OBJECTS *jro;

    jro = (OBJECTS *)(self->gc.jros+sizeof(OBJECTS)*handle);
    return (--jro->count);
    }

METHOD UINT gc_gc_inc_access(PR_GC *self, UINT handle)
/*  Increment access count of object 'handle'  */
    {
    OBJECTS *jro;

    jro = (OBJECTS *)(self->gc.jros+sizeof(OBJECTS)*handle);
    return (++jro->count);
    }

